/*
 * transfer.c - transfer core
 *
 * Written by
 *  Christian Vogelgsang <chris@vogelgsang.org>
 *
 * This file is part of dtv2ser.
 * See README for copyright notice.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 *  02111-1307  USA.
 *
 */

#include "global.h"

#include <util/crc16.h>

#include "transfer.h"
#include "display.h"
#include "util.h"
#include "timer.h"
#include "param.h"

#define min(a,b) ((a<b)?(a):(b))

// the transfer state
transfer_state_t transfer_state;

// the internal dtv transfer state
static dtv_transfer_state_t dtv_state;

// ----- main transfer loop -----

static u08 transfer_begin(u08 mode,u32 length,
                          host_transfer_funcs_t *host_funcs)
{
  // reset transfer state
  transfer_state.length  = 0;
  transfer_state.ms_time = 0;
  transfer_state.result  = TRANSFER_OK;

  // begin host transfer
  u08 result = host_funcs->begin_transfer(length);
  if(result!=TRANSFER_OK) {
    transfer_state.result = result;
    return result;
  }
  
  // reset dtv state
  dtv_state.transfer_length = 0;
  dtv_state.mode = mode;

  led_transmit_on();  
  return TRANSFER_OK;
}

static u08 transfer_end(u08 result,u32 total_length,u16 ms_time,
                        host_transfer_funcs_t *host_funcs)
{  
  led_transmit_off();
  
  // end host transfer
  result = host_funcs->end_transfer(result);
  
  // set transfer result
  transfer_state.length  = total_length;
  transfer_state.ms_time = ms_time;
  transfer_state.result  = result;
  
  return result;
}

u08 transfer_mem(u08 mode,u32 base,u32 length,u16 block_size,
                 host_transfer_funcs_t *host_funcs,
                 dtv_transfer_block_func_t dtv_func)
{
  u08 result = transfer_begin(mode,length,host_funcs);
  if(result!=TRANSFER_OK)
    return result;

  // start timer
  timer_10ms = 0;
    
  // block copy loop
  u32 total_length = 0;
  u08 toggle = 1;
  while(length) {
    u16 offset = (u16)(base & 0x3fff);
    u16 max_len = 0x4000 - offset;
    u16 block_len = min(length,block_size);
    u16 len = min(block_len,max_len);

    dtv_state.offset = offset;
    dtv_state.bank   = (u08)(base >> 14);
    dtv_state.length = len;
    dtv_state.crc16  = 0xffff;
    dtv_state.base   = (u16)(base & 0xffff); // for BOOT
    
    // begin host block
    result = host_funcs->begin_block(len);
    if(result!=TRANSFER_OK)
      break;
    
    // call dtv func to transfer a single block
    // (calls host_funcs transfer_byte)
    result = dtv_func(&dtv_state,host_funcs->transfer_byte);

    // host end block
    result = host_funcs->end_block(result,dtv_state.crc16);
    if(result!=TRANSFER_OK)
      break;

    // update 
    u16 transfer_length = dtv_state.transfer_length;
    base   += transfer_length;
    length -= transfer_length;
    total_length += transfer_length;

    // toggle led
    toggle = toggle^1;
    if(toggle)
      led_transmit_on();
    else
      led_transmit_off();

    // break if not ok
    if(result!=TRANSFER_OK)
      break;
  }
  
  return transfer_end(result,
                      total_length,
                      timer_10ms,
                      host_funcs);
}

u08 transfer_mem_block(u08 mode,u08 bank,u16 offset,u16 length,
                       host_transfer_funcs_t *host_funcs,
                       dtv_transfer_block_func_t dtv_func)
{
  u08 result = transfer_begin(mode,length,host_funcs);
  if(result!=TRANSFER_OK)
    return result;

  dtv_state.bank   = bank;
  dtv_state.offset = offset;
  dtv_state.length = length;

  // trigger a single block transfer
  result = dtv_func(&dtv_state,host_funcs->transfer_byte);
  
  return transfer_end(result,
                      dtv_state.transfer_length,
                      0,
                      host_funcs);
}

// ----- diagnose transfer functions ----------------------------------------
#ifdef USE_DIAGNOSE

u08 diagnose_dtv_send_block(dtv_transfer_state_t *state,
                            host_transfer_byte_func_t transfer_byte)
{
  u08 result;
  u16 length = state->length;

  u16 pos = 0;
  u16 crc16 = state->crc16;
  while(length) {
    // transfer byte from host
    u08 data;
    result = transfer_byte(&data);
    if(result!=TRANSFER_OK)
      break;
      
    // update checksum
    crc16 = _crc16_update(crc16,data);
    
    length--;
    pos++;
  }
  state->transfer_length = pos;
  state->crc16 = crc16;
  
  return TRANSFER_OK;
}

u08 diagnose_dtv_recv_block(dtv_transfer_state_t *state,
                            host_transfer_byte_func_t transfer_byte)
{
  u08 result;
  u16 length = state->length;
  u08 pattern = PARAM_BYTE(PARAM_BYTE_DIAGNOSE_PATTERN);
  
  u16 pos = 0;
  u16 crc16 = state->crc16;
  while(length) {
    // transfer pattern byte to host
    result = transfer_byte(&pattern);
    if(result!=TRANSFER_OK)
      break;
    
    // update checksum
    crc16 = _crc16_update(crc16,pattern);
    
    pos++;
    length--;
  }
  state->transfer_length = pos;
  state->crc16 = crc16;
  
  return TRANSFER_OK;
}

// ----- diagnose host transfer funcs -----

u08 diagnose_host_mode = 0; // 0=read, 1=write, 2=verify

static u08 diagnose_begin_transfer(u32 length)
{ return TRANSFER_OK; }

static u08 diagnose_end_transfer(u08 lastStatus)
{ return lastStatus; }

static u08 diagnose_begin_block(u16 length)
{ return TRANSFER_OK; }

static u08 diagnose_end_block(u08 lastStatus,u16 crc16)
{ return lastStatus; }

static u08 diagnose_transfer_byte(u08 *data)
{ 
  u08 valid = PARAM_BYTE(PARAM_BYTE_DIAGNOSE_PATTERN);
  if(diagnose_host_mode==DIAGNOSE_HOST_MODE_READ)
    *data = valid;
  else if(diagnose_host_mode==DIAGNOSE_HOST_MODE_VERIFY)
    if(*data != valid)
      return TRANSFER_ERROR_VERIFY_MISMATCH;
  // do nothing in DIAGNOSE_HOST_MODE_WRITE
  return TRANSFER_OK; 
}

host_transfer_funcs_t diagnose_host_transfer_funcs =
{
  .begin_transfer = diagnose_begin_transfer,
  .end_transfer   = diagnose_end_transfer,
  .begin_block    = diagnose_begin_block,
  .end_block      = diagnose_end_block,
  .transfer_byte  = diagnose_transfer_byte
};

#endif // USE_DIAGNOSE

