/*
 * uart.c - serial hw routines
 *
 * Written by
 *  Christian Vogelgsang <chris@vogelgsang.org>
 *
 * This file is part of dtv2ser.
 * See README for copyright notice.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 *  02111-1307  USA.
 *
 */

#include "global.h"

#include <avr/io.h>
#include <avr/interrupt.h>

#include "uart.h"
#include "timer.h"
#include "util.h"
#include "param.h"
#include "display.h"

#ifdef UBRR0H

// for atmeg644
#define UBRRH  UBRR0H
#define UBRRL  UBRR0L
#define UCSRA  UCSR0A
#define UCSRB  UCSR0B
#define UCSRC  UCSR0C
#define UDRE   UDRE0
#define UDR    UDR0

#define RXC    RXC0
#define TXC    TXC0
#define DOR    DOR0
#define PE     UPE0

#endif

// calc ubbr from baud rate
#define UART_UBRR   F_CPU/16/UART_BAUD-1

#define UART_RX_BUF_SIZE 16
static volatile u08 uart_rx_buf[UART_RX_BUF_SIZE];
static volatile u08 uart_rx_start = 0;
static volatile u08 uart_rx_end = 0;

void uart_init(void) 
{
  cli();

  // baud rate
  UBRRH = (u08)((UART_UBRR)>>8);
  UBRRL = (u08)((UART_UBRR)&0xff);

  UCSRB = 0x98; // 0x18  enable tranceiver and transmitter, RX interrupt
  UCSRC = 0x86; // 0x86 -> use UCSRC, 8 bit, 1 stop, no parity, asynch. mode

  sei();
}

// receiver interrupt
ISR(USART_RXC_vect)
{
  u08 data = UDR;
  uart_rx_buf[uart_rx_end] = data;

  uart_rx_end++;
  if(uart_rx_end == UART_RX_BUF_SIZE)
    uart_rx_end = 0;

//#define CHECK_UART_ERROR
#ifdef CHECK_UART_ERROR
  // overrun?
  if(uart_rx_end==uart_rx_start) {
    led_error_on();
  }
  
  // uart error?
  u08 status = UCSRA;
  if ((status & (_BV(FE)|_BV(DOR)|_BV(PE))) != 0) {
    led_error_on();
  }
#endif
}

u08 uart_data_available(void)
{
  return uart_rx_start != uart_rx_end;
}

void uart_clear_buffer(void) 
{
  uart_rx_start = 0;
  uart_rx_end = 0;
}

static void uart_read_buffer(u08 *data)
{
  *data = uart_rx_buf[uart_rx_start];
  uart_rx_start++;
  if(uart_rx_start==UART_RX_BUF_SIZE)
    uart_rx_start = 0;  
}

// read/write without RTS/CTS

u08 uart_read(u08 *data) {
  // read for buffe to be filled
  timer_100us = 0;
  u16 timeout = PARAM_WORD(PARAM_WORD_SERIAL_READ_AVAIL_TIMEOUT);
  while(uart_rx_start==uart_rx_end) {
    if (timer_100us > timeout) {
      return 0;
    }
  }

  // read buffer
  uart_read_buffer(data);
  return 1;
}

u08 uart_send(u08 data)
{
  // wait for transmitter to become ready
  timer_100us = 0;
  u16 timeout = PARAM_WORD(PARAM_WORD_SERIAL_SEND_READY_TIMEOUT);
  while(!( UCSRA & (1<<UDRE))) {
    if (timer_100us > timeout) {
      return 0;
    }
  }
  UDR = data;
  return 1;
}

// read/write with RTS/CTS

u08 uart_read_with_cts(u08 *data)
{
  // some data in buffer?
  if(uart_rx_start!=uart_rx_end) {
    uart_read_buffer(data);
    return 1;
  }
  
  // request new data
  uart_set_cts(1);
  u08 status = uart_read(data);
  uart_set_cts(0);
  return status;
}

u08 uart_send_with_rts(u08 data)
{
  // wait for rts with timeout
  timer_100us = 0;
  u16 rts_timeout = PARAM_WORD(PARAM_WORD_SERIAL_RTS_TIMEOUT);

#ifdef SHOW_WAIT_RTS
  led_ready_on();
#endif
  while(uart_get_rts()==0) {
    if(timer_100us > rts_timeout) {
#ifdef SHOW_WAIT_RTS      
      led_ready_off();
#endif
      return 0;
    }
  }
#ifdef SHOW_WAIT_RTS
  led_ready_off();
#endif
  return uart_send(data);
}

// ----- Tools -----

u08 uart_send_string(u08 *str)
{
  while(*str) {
    if(!uart_send(*str))
      return 0;
    str++;
  }
  return 1;
}

u08 uart_send_data(u08 *data,u08 len)
{
  for(u08 i=0;i<len;i++) {
    if(!uart_send(data[i]))
      return 0;
  }
  return 1;
}

u08 uart_send_crlf(void)
{
  return uart_send_string((u08 *)"\r\n");
}

static u08 buf[6];

u08 uart_send_hex_byte_crlf(u08 data)
{
  byte_to_hex(data,buf);
  if(uart_send_data(buf,2))
    return uart_send_crlf();
  else
    return 0;
}

u08 uart_send_hex_word_crlf(u16 data)
{
  word_to_hex(data,buf);
  if(uart_send_data(buf,4))
    return uart_send_crlf();
  else
    return 0;
}

u08 uart_send_hex_dword6_crlf(u32 data)
{
  dword_to_hex6(data,buf);
  if(uart_send_data(buf,6))
    return uart_send_crlf();
  else
    return 0;
}

